
#include <fftw3.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "match-with-dont-cares.h"

void printUsageAndExit(void){
	fprintf(stderr,"Usage: matcher text pattern n m [wild card char] \n");
	fprintf(stderr,"       m must be less than or equal to n, and both must be strictly positive.\n");
	fprintf(stderr,"       Default wild card is ? e.g. matcher text pattern n m ?\n");
	exit(0);
}

void matchWithFFT(char *text, char *pattern,char wildCardChar, int n, int m,int indexOffset){

	int i;

	/* Convert text to an array of doubles; and double the degree bound.
	 * Compute squares and cubes of each element as we go */
	double *t, *tsquared, *tcubed;
	t = fftw_malloc(sizeof(double) * n*2);
	tsquared = fftw_malloc(sizeof(double) * n*2);
	tcubed = fftw_malloc(sizeof(double) * n* 2);

	for(i=0; i<n; i++){
		t[i] = (text[i]==wildCardChar) ? 0.0 : (double) text[i];
		tsquared[i] = t[i] * t[i];
		tcubed[i] = t[i] * tsquared[i];
	}

	for(;i<2*n;i++){
		t[i] = 0.0;
		tsquared[i] = 0.0;
		tcubed[i] = 0.0;
	}

	/* Convert pattern to an array of doubles and pad to the same size as the text
	 * Reverse pattern as we go
	 *Compute squares and cubes of each element as we go */

	double *p,*psquared, *pcubed;
	p = fftw_malloc(sizeof(double) * n*2);
	psquared = fftw_malloc(sizeof(double) * n*2);
	pcubed= fftw_malloc(sizeof(double) * n*2);

	for(i=0; i<m; i++){
		p[i] = (pattern[m-i-1]==wildCardChar) ? 0.0 : (double) pattern[m-i-1];
		if(p[m-i-1]==wildCardChar){
			printf("index %d in pattern is wild card character\n",m-i-1);
		}
		psquared[i] = p[i] * p[i];
		pcubed[i] = p[i] * psquared[i];
	}

	for(;i<2*n;i++){
		p[i] = 0.0;
		psquared[i] = 0.0;
		pcubed[i] = 0.0;
	}

	/*Calculate the DFT of the text, square of text and cube of text*/
	fftw_complex *DFTofText, *DFTofTextSquared, *DFTofTextCubed;
	DFTofText = fftw_malloc(sizeof(fftw_complex) * n*2);
	DFTofTextSquared = fftw_malloc(sizeof(fftw_complex) * n*2);
	DFTofTextCubed = fftw_malloc(sizeof(fftw_complex) * n*2);
	doDFT(n*2,t,DFTofText);
	doDFT(n*2,tsquared,DFTofTextSquared);
	doDFT(n*2,tcubed,DFTofTextCubed);

	//Free the original text values
	fftw_free(t);
	fftw_free(tsquared);
	fftw_free(tcubed);

	/*Calculate the DFT of the pattern, square of pattern and cube of pattern */

	fftw_complex *DFTofPattern, *DFTofPatternSquared, *DFTofPatternCubed;
	DFTofPattern = fftw_malloc(sizeof(fftw_complex) * n*2);
	DFTofPatternSquared = fftw_malloc(sizeof(fftw_complex) * n*2);
	DFTofPatternCubed = fftw_malloc(sizeof(fftw_complex) * n*2);
	doDFT(n*2,p,DFTofPattern);
	doDFT(n*2,psquared,DFTofPatternSquared);
	doDFT(n*2,pcubed,DFTofPatternCubed);

	//Free the original pattern values
 	fftw_free(p);
 	fftw_free(psquared);
 	fftw_free(pcubed);

	/* Multiply the point representations*/

	fftw_complex *pCubedTimesT_PR = fftw_malloc(sizeof(fftw_complex) * n * 2);
	fftw_complex *pSquaredTimesTSquared_PR = fftw_malloc(sizeof(fftw_complex) * n * 2);
	fftw_complex *pTimesTCubed_PR = fftw_malloc(sizeof(fftw_complex) * n * 2);

	for(i=0;i<n*2;i++){
		pCubedTimesT_PR[i][0] = DFTofPatternCubed[i][0] * DFTofText[i][0] - DFTofPatternCubed[i][1] * DFTofText[i][1];
		pCubedTimesT_PR[i][1] = DFTofPatternCubed[i][0] * DFTofText[i][1] + DFTofPatternCubed[i][1] * DFTofText[i][0];

		pSquaredTimesTSquared_PR[i][0] = DFTofPatternSquared[i][0] * DFTofTextSquared[i][0] - DFTofPatternSquared[i][1] * DFTofTextSquared[i][1];
		pSquaredTimesTSquared_PR[i][1] = DFTofPatternSquared[i][0] * DFTofTextSquared[i][1] + DFTofPatternSquared[i][1] * DFTofTextSquared[i][0];

		pTimesTCubed_PR[i][0] = DFTofPattern[i][0] * DFTofTextCubed[i][0] - DFTofPattern[i][1] * DFTofTextCubed[i][1];
		pTimesTCubed_PR[i][1] = DFTofPattern[i][0] * DFTofTextCubed[i][1] + DFTofPattern[i][1] * DFTofTextCubed[i][0];

	}

	//Free the DFTs
 	fftw_free(DFTofText);
 	fftw_free(DFTofTextSquared);
 	fftw_free(DFTofTextCubed);
 	fftw_free(DFTofPattern);
 	fftw_free(DFTofPatternSquared);
 	fftw_free(DFTofPatternCubed);

	/* Convert back to a coefficient representation */
	double *pCubedTimesT, *pSquaredTimesTSquared, *pTimesTCubed;
	pCubedTimesT = fftw_malloc(sizeof(double) * n*2);
	pSquaredTimesTSquared = fftw_malloc(sizeof(double) * n*2);
	pTimesTCubed= fftw_malloc(sizeof(double) * n*2);
	doInverseDFT(n*2,pCubedTimesT_PR,pCubedTimesT);
	doInverseDFT(n*2,pSquaredTimesTSquared_PR,pSquaredTimesTSquared);
	doInverseDFT(n*2,pTimesTCubed_PR,pTimesTCubed);

	printMatches(pCubedTimesT,pSquaredTimesTSquared,pTimesTCubed,n,m,indexOffset);

	//Free the rest of the allocated values

 	fftw_free(pCubedTimesT_PR);
 	fftw_free(pSquaredTimesTSquared_PR);
 	fftw_free(pTimesTCubed_PR);

 	fftw_free(pCubedTimesT);
 	fftw_free(pSquaredTimesTSquared);
 	fftw_free(pTimesTCubed);
}

void doDFT(int n,double *in, fftw_complex * out){
	fftw_plan plan;
	plan = fftw_plan_dft_r2c_1d(n,in,out,FFTW_ESTIMATE);
	fftw_execute(plan);
}

void doInverseDFT(int n,fftw_complex *in, double * out){
	fftw_plan plan;
	plan = fftw_plan_dft_c2r_1d(n,in,out,FFTW_ESTIMATE);
	fftw_execute(plan);
	fftw_destroy_plan(plan);
}

void printMatches(double *pCubedTimesT,double *pSquaredTimesTSquared,double *pTimesTCubed,int n,int m, int indexOffset){
	int i;

	//divide each by n*2 as fftw does an un-normalised transform
	for(i=0;i<=n-m;i++){
		printf("testing index %d, offset value %d\n",i,i+indexOffset);
		double result = pCubedTimesT[m+i-1]/(n*2) - (pSquaredTimesTSquared[m+i-1]/n) + pTimesTCubed[m+i-1]/(n*2);
		if(fabs(result) == 0.0){
			printf("Match at index %d\n",i+indexOffset);
			//Work required to ascertain a suitable threshold to accept input
		}else if(fabs(result) <= 0.00001){
			printf("Near match at index %d with value %f\n",i+indexOffset,result);
		}
	}
}
